import sys

#print ("sys")
#sys.path.insert(0,'C:\\Users\\User\\Documents\\GitHub\\pytec\\src\\pytec')

#for p in sys.path:
#    print(p)


from ionotec import tec

#tec.st.create_output("output")

#sys.exit()

data="C:\\Users\\User\\Documents\\TEC\\data\\"

f_obs=[data+"2022\\285\\antc2850.22o",data+"2022\\286\\antc2860.22o",data+"2022\\287\\antc2870.22o"]
f_nav=[data+"2022\\285\\brdc2850.22n",data+"2022\\286\\brdc2860.22n",data+"2022\\287\\brdc2870.22n"]
f_bias=data+"2022\\285\\P1C12210.DCB"

station_tec = tec.tec(f_obs,f_nav,f_bias)
station_tec.compute_vtec()

from pytec import graph

graph.plot_station(station_tec.df_obs,"output/antc",mozaic=False)



import matplotlib.pyplot as plt
import pandas as pd
import sys


def plot_station(df_station,png_file_name,mozaic=False):

    list_sats = df_station["sv"].unique()

    if mozaic:
        fig, axs = plt.subplots(4,2,figsize=(14,10),sharex=True,sharey=True)

        i,j=0,0
        n_sat = 0
        n=0
        N=0
        for sat in list_sats:
            i = int(n/2)
            j = n%2
            df_sat = df_station[df_station["sv"]==sat]
            axs[i,j].plot(df_sat["VTEC"],'b.',markersize=1)
            axsx = axs[i,j].twinx()
            axsx.plot(df_sat["elevation"]*180/3.1415926535,'r.',markersize=1,alpha=0.5)
            axsx.set_ylim([0,90])
            axs[i,j].set_title(sat)
            axs[i,j].grid(True)
            axs[i,j].set_ylabel("VTEC")
            axsx.set_ylabel("elevation")
            axsx.spines['right'].set_color('red')
            axsx.tick_params(axis='y', colors='red')
            axsx.yaxis.label.set_color('red')

            n = n+1
            n_sat = n_sat+1

            if n==8:
                n=0
                plt.savefig(png_file_name+"-"+str(N)+".png",bbox_inches='tight')
                plt.close()
                fig, axs = plt.subplots(4,2,figsize=(14,10),sharex=True,sharey=True)
                N = N+1
        if n!=0:
            plt.savefig(png_file_name+"-"+str(N)+".png",bbox_inches='tight')
            plt.close()
            

    else:
        fig, axs = plt.subplots(1,figsize=(10,7),sharex=True,sharey=True)
        
        i,j=0,0
        n_sat = 0
        n=0
        N=0
        min_t, max_t = min(df_station.index), max(df_station.index)
        for sat in list_sats:
            df_sat = df_station[df_station["sv"]==sat]
            axs.scatter(df_sat.index,df_sat["VTEC"].values,s=1,c=df_sat["elevation"]*2/3.1415926535,cmap="YlGnBu")

            axs.set_title(png_file_name.split("/")[-1]+" "+min_t.strftime("%d/%m/%Y %H:%M:%S") + " to " + max_t.strftime("%d/%m/%Y	 %H:%M:%S") )
            axs.grid(True)
            axs.set_ylabel("VTEC")

        print (png_file_name)
        plt.savefig(png_file_name+".png",bbox_inches='tight')



#############################################################################3
#
# MIT License
#
# Copyright (c) 2021 Sylvain Blunier
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#
#############################################################################3

import os,sys,shutil

#import sys
import numpy as np
import georinex as gr
import datetime
import math
import numpy as np
#import urllib.request
import pymap3d as pm
#from unlzw import unlzw

from warnings import simplefilter
simplefilter(action='ignore', category=FutureWarning)


import pandas as pd

#import random
#import time

#import matplotlib.pyplot as plt
from . import stations as st

from os import listdir
from os.path import isfile, join

#Ionosphere altitude
Alt_i = 300000
# Earth Radius
R_E = 6370000
omega_e = 7.292115e-5
GM = 3.986005e14
C20 = -1082.63e-6
a_major_ellipsoid_axis = 6378136
WG84_to_PZ90=np.matrix([[1,-0.33/3600.0,0],[0.33/3600.0,1,0],[0,0,1]])


#root_dir = "/media/sblunier/CSN_TEC/"
#root_dir = os.getenv("ROOT_RINEX")
#root_dir = "/home/sblunier/Work/spaceweather/tec/"
#dir_rinex_nav = "/home/sblunier/Work/spaceweather/data/GNSS/"
#dir_rinex_nav = root_dir+"2020/349/"
#dir_sat_bias = "/home/sblunier/Work/spaceweather/data/GNSS/CODE/"
#dir_sat_bias = dir_rinex_nav

def F1(pos,v):
	return np.array(v)

def F2(pos,v,a):
	r=np.linalg.norm(pos)
	xpp = -GM*pos[0]/r**3 + 1.5*C20*GM*a_major_ellipsoid_axis**2 * pos[0] * abs(1-5*pos[2]**2/r**2) / r**5  + omega_e**2*pos[0] + 2*omega_e*v[1]+a[0]
	ypp = -GM*pos[1]/r**3 + 1.5*C20*GM*a_major_ellipsoid_axis**2 * pos[1] * abs(1-5*pos[2]**2/r**2) / r**5 + omega_e**2*pos[1] - 2*omega_e*v[0]+a[1]
	zpp = -GM*pos[2]/r**3 + 1.5*C20*GM*a_major_ellipsoid_axis**2 * pos[2] * abs(3-5*pos[2]**2/r**2) / r**5
	return np.array([xpp,ypp,zpp])

def getDateFromSat(weekNumber,seconds):
	date_ref = datetime.datetime(1980,1,6,0,0,0)
	days = weekNumber*7
	sec = int(seconds) #- 604800
	while sec > 86399:
		days = days + 1
		sec -= 86400
	date = date_ref + datetime.timedelta(days=days) + datetime.timedelta(seconds=sec)
	return date

def gps_nav_to_XYZ(data,date):
    Toe = data['Toe']
    TGD = data['TGD']
    IDOT = data['IDOT']
    IODC = data['IODC']
    GPSWeek = data['GPSWeek']
    TransTime = data['TransTime']
    #clock correction biais
    clockBiais = data['SVclockBias']
    #clock correction drift
    clockDrift = data['SVclockDrift']
	#clock correction drift rate
    clockDriftRate = data['SVclockDriftRate']
	#Square root of the semi-major axis m^{1/2}
    sqrtA =  data['sqrtA']
	#Eccentricity
    Eccentricity =  data['Eccentricity']
	#Inclination angle at reference time
    Io =  data['Io']
	#Longitude of ascending node at reference time
    Omega0 =  data['Omega0']
	#Argument of perigee (semicircles)
    omega = data['omega']
	#Mean anomaly at reference time (semicircles)
    M0 = data['M0']
	#Mean motion difference from computed value
    DeltaN = data['DeltaN']
	#Rate of change of right ascension
    OmegaDot = data['OmegaDot']
	#Rate of change of inclination
    IDOT = data['IDOT']
	#Amplitude of the sine harmonic correction term to the argument of latitude (rad)
    Cus = data['Cus']
	#Amplitude of the cosine harmonic correction term to the argument of latitude (rad)
    Cuc = data['Cuc']
	#Amplitude of the sine harmonic correction term to the angle of inclination (rad)
    Cis = data['Cis']
	#Amplitude of the cosine harmonic correction term to the angle of inclination (rad)
    Cic = data['Cic']
	#Amplitude of the sine harmonic correction term to the orbit radius (m)
    Crs = data['Crs']
	#Amplitude of the cosine harmonic correction term to the orbit radius (m)
    Crc = data['Crc']

    d0 = getDateFromSat(GPSWeek,Toe)

    tk = (date-d0).seconds

    n0 = math.sqrt(GM)/sqrtA**3
    n = n0 + DeltaN

    Mk = M0 + n*tk
    Ek = Mk
    for i in range(3): Ek = Mk + Eccentricity*math.sin(Ek)

    sin_vk = math.sqrt(1-Eccentricity**2)*math.sin(Ek) / (1.0-Eccentricity*math.cos(Ek))
    cos_vk = (math.cos(Ek)-Eccentricity)/(1.0-Eccentricity*math.cos(Ek))
    vk=0
    if sin_vk>0: vk = math.acos(cos_vk)
    else: vk = -math.acos(cos_vk)

    Phik = vk + omega
    delta_uk = Cuc * math.cos(2*Phik) + Cus * math.sin(2*Phik)
    delta_rk = Crc * math.cos(2*Phik) + Crs * math.sin(2*Phik)
    delta_ik = Cic * math.cos(2*Phik) + Cis * math.sin(2*Phik)

    uk = Phik + delta_uk
    rk = sqrtA**2 * (1 - Eccentricity * math.cos(Ek)) + delta_rk
    ik = Io + IDOT*tk + delta_ik

    Xk_p = rk * math.cos(uk)
    Yk_p = rk * math.sin(uk)

    Omegak = Omega0 + (OmegaDot-omega_e)*tk - omega_e * Toe

    Xk = Xk_p * math.cos(Omegak) - Yk_p * math.sin(Omegak)* math.cos(ik)
    Yk = Xk_p * math.sin(Omegak) + Yk_p * math.cos(Omegak)* math.cos(ik)
    Zk = Yk_p * math.sin(ik)
    return [Xk,Yk,Zk]

dir_sats = "sats/"

class gnss:
    '''
        gnss Class to perform eficiently the position and elevation of the
        satellites
        for now it only works on GPS constellation, future version should deal
        with GLONASS
    '''
    def __init__(self,f_nav=[],resolution=60):
     
        self.gps_dir = st.root_dir + "GPS/"
        if not os.path.exists(self.gps_dir):
            try: os.mkdir(self.gps_dir)
            except OSError as e:
                if e.errno!=17: print ("FAIL creation of directory "+self.gps_dir, e )
            else: print ("Successfully created the directory "+self.gps_dir)

        self.list_f_rinex_nav = f_nav
        self.dict_df_pos = {}
        self.df_pos = pd.DataFrame()
        self.resolution = resolution
        self.compute_position(self.list_f_rinex_nav)


    # Function that loads the satellite of a specific year, files must exist, otherwise load empty
    def load_sats(self,d_in,d_out):
    
        year = d_in.year
        first = True
         
        for i in range(1,33):
            #Name of the satellite
            sat = ""
            if i<10: sat = "G0"+str(i)
            else: sat = "G"+str(i)

            sat_file = self.gps_dir + str(year) + "/" + sat + ".feather"

            if os.path.exists(sat_file):
                self.dict_df_pos[sat] = pd.read_feather(sat_file)
                self.dict_df_pos[sat].set_index("time",inplace=True)
                self.dict_df_pos[sat].index = pd.to_datetime(self.dict_df_pos[sat].index)
                mask = (self.dict_df_pos[sat].index>=d_in) & (self.dict_df_pos[sat].index<=d_out)
                self.dict_df_pos[sat]=self.dict_df_pos[sat].loc[mask]
                if first:
                    self.df_pos = self.dict_df_pos[sat]
                    self.df_pos["sv"] = sat
                    first = False
                else:
                    df_inter = self.dict_df_pos[sat]
                    df_inter["sv"] = sat
                    self.df_pos = pd.concat([self.df_pos,df_inter])
            else:
                self.dict_df_pos[sat] = pd.DataFrame()

       
    # Function that compute the position of the satellites from the navigation file 
    #   with the resolution informed when instanciating the gnss object
    def compute_position(self,form='feather'):
    
        # Columns that need to be used
        list_cols = ["Toe","TGD","IDOT","IODC","GPSWeek","TransTime","SVclockBias","SVclockDrift",\
        "SVclockDriftRate","sqrtA","Eccentricity","Io","Omega0","omega","M0","DeltaN","OmegaDot","Cus","Cuc",\
        "Cis","Crs","Crc","Cic"] 

        for f_rinex_nav in self.list_f_rinex_nav:

            rinex_doy = int(f_rinex_nav[-8:-5])
            rinex_year = int("20"+f_rinex_nav[-3:-1])

            # Construct path of satellite files and create directory if does not exists
            year_gpsdir = self.gps_dir + str(rinex_year)
            if not os.path.exists(year_gpsdir):
                try: os.mkdir(year_gpsdir)
                except OSError as e:
                    if e.errno!=17: print ("FAIL creation of directory "+year_gpsdir, e )
                    else: print ("Successfully created the directory "+year_gpsdir)    
            
            #Annex file reporting doy that have been considered for each satellite
            f_doy_reported = year_gpsdir + "/reported_doy.feather"
            
            
            if not os.path.exists(f_doy_reported):
                dict_doy_reported = {"sat":[],"res":[]}
                for doy in range(366):
                    dict_doy_reported[str(doy)] = []
                for i in range(1,33):
                    #Name of the satellite
                    sat = ""
                    if i<10: sat = "G0"+str(i)
                    else: sat = "G"+str(i)
                    dict_doy_reported["sat"].append(sat)
                    dict_doy_reported["res"].append(self.resolution )
                    for doy in range(366):
                        dict_doy_reported[str(doy)].append(0)
                pd.DataFrame(dict_doy_reported).to_feather(f_doy_reported)      

            # Load report of already processed files in order to know which doy have been processed 
            #   for each satellite and with which resolution
            df_doy_reported = pd.read_feather(f_doy_reported)
            df_doy_reported.set_index("sat",inplace=True)
           
            # Prepare time list that will be used
            t_base = datetime.datetime(rinex_year,1,1,0,0,0,0)+datetime.timedelta(days=rinex_doy-1)
            t=t_base
            time_list = []
            while t<t_base+datetime.timedelta(days=1):
                time_list.append(t)
                t += datetime.timedelta(seconds=self.resolution)

        
            # Load provided navigation file
            try:
                nav = gr.load(f_rinex_nav)
                df_nav = nav.to_dataframe()
            except ValueError:
                print ("COuld not load file", f_rinex_nav)
                continue
                

            # Make if friendly to use with dataframe
            df_nav.index.set_names(["time","sv"],inplace=True)
            df_nav.reset_index(level=["sv"],inplace=True)
            df_nav.dropna(inplace=True)
       
            # Iter over all the satellites
            for i in range(1,33):
                #Name of the satellite
                sat = ""
                if i<10: sat = "G0"+str(i)
                else: sat = "G"+str(i)

                # flag to know if position of satellite has already been computed for this doy
                is_doy_reported = df_doy_reported[str(rinex_doy)].loc[sat]==1
            
                # flag to know if the resolution is fine
                is_res_enough = df_doy_reported["res"].loc[sat]<=self.resolution

                # If this doy has been reported with enough time-resolution, got for next satellite
                if is_doy_reported and is_res_enough: continue
            
                # import data of this satellite
                df_sat_new = df_nav[df_nav["sv"]==sat][list_cols]          
            
                # If no data has been found for this satellite, go for next one without creating the feather
                if len(df_sat_new)==0:
                    print ("No position for satellite",sat)
                    continue
            
                # File correponding to satellite of interest
                csv_nav_sat = year_gpsdir + "/" + sat + ".feather"
            
                # Intermediate dictionnary intended to contain data of the satellite under process 
                dict_sat_pos = {"time":[],"X":[],"Y":[],"Z":[]}
            
                # Iterator of time list of positions to be computed
                i_time_list = 0
                # Get first time
                date = time_list[i_time_list]
            
                # Iter over the times available in the navigation file
                for t_nav, row in df_sat_new.iterrows():
                    # Calculate position for each time in time_list that are before the next available position information
                    while date < t_nav and i_time_list<len(time_list):
                        sat_pos = gps_nav_to_XYZ(row,date)
                        dict_sat_pos["time"].append(date)
                        dict_sat_pos["X"].append(sat_pos[0])
                        dict_sat_pos["Y"].append(sat_pos[1])
                        dict_sat_pos["Z"].append(sat_pos[2])
                        i_time_list += 1
                        if i_time_list==len(time_list): break
                        date = time_list[i_time_list]
                last_row = df_sat_new.iloc[-1]
            
                # Case novigation data does not provide position until 00:00, use last available date
                while i_time_list<len(time_list):
                    sat_pos = gps_nav_to_XYZ(last_row,date)
                    dict_sat_pos["time"].append(date)
                    dict_sat_pos["X"].append(sat_pos[0])
                    dict_sat_pos["Y"].append(sat_pos[1])
                    dict_sat_pos["Z"].append(sat_pos[2])
                    i_time_list += 1
                    if i_time_list==len(time_list): break
                    date = time_list[i_time_list]
            
                df_sat_nav = pd.DataFrame(dict_sat_pos)
                df_sat_nav.set_index("time",inplace=True)
            
                # Check if feather file is not already there in order not to use navigation file
                if os.path.exists(csv_nav_sat):
                    #print (csv_nav_sat, "file exist, concat with new results")
                    df_sat_old = pd.read_feather(csv_nav_sat) 
                    df_sat_old.set_index("time",inplace=True)
                    df_sat_old.index = pd.to_datetime(df_sat_old.index)
                    df_sat_nav = pd.concat([df_sat_nav,df_sat_old])
                          
                df_sat_nav.reset_index(inplace=True)
                df_sat_nav.to_feather(csv_nav_sat)
            
                # Report that this doy has been treated with the corresponding resolution to avoid reprocessing
                df_doy_reported[str(rinex_doy)].loc[sat]=1
                df_doy_reported["res"].loc[sat]=self.resolution
        
            # Update report file
            df_doy_reported.reset_index(inplace=True)
            df_doy_reported.to_feather(f_doy_reported)
            

        
 
    
    def getElevation(self,df,pos_antena):

        norm_antena = np.linalg.norm(pos_antena)
        
        dfdif = pd.DataFrame()   
        dfdif['Xdif'] = self.df_pos["X"]-pos_antena[0]
        dfdif['Ydif'] = self.df_pos["Y"]-pos_antena[1]
        dfdif['Zdif'] = self.df_pos["Z"]-pos_antena[2]

        nv = pos_antena[0]*dfdif["Xdif"] + pos_antena[1]*dfdif["Ydif"] + pos_antena[2]*dfdif["Zdif"]
        norm = norm_antena * np.linalg.norm(dfdif[['Xdif','Ydif','Zdif']].values,axis=1) 
        
        sin_phi = nv/norm
        self.df_pos["elevation"] = np.arcsin(sin_phi)
        return pd.merge(df,self.df_pos,how='left',on=['sv','time']) 


    def getPiercingPoint(self,df,pos_antena,h=400000):
        R_I = R_E + h
        df_inter = pd.DataFrame()

        #### Calculation of the position of the intersection of Satellite-Antena line with ionosphere which is a second degree equation
        df_inter["Xas"] = df["X"]-pos_antena[0]
        df_inter["Yas"] = df["Y"]-pos_antena[1]
        df_inter["Zas"] = df["Z"]-pos_antena[2]
        
        # return elevation of satellite given the ID os the satellite, the position of the antenna.
        df_inter["AS_squared"] = df_inter["Xas"]**2 + df_inter["Yas"]**2 + df_inter["Zas"]**2
        #Parameters of the polynomial to solve
        df_inter["b_poly"] = 2*df_inter["Xas"]*(pos_antena[0]*df_inter["Xas"] + pos_antena[1]*df_inter["Yas"] + pos_antena[2]*df_inter["Zas"])
        df_inter["c_poly"] = df_inter["Xas"]**2 * ( pos_antena[0]**2 + pos_antena[1]**2+ pos_antena[2]**2 - R_I**2 )
        df_inter["Delta"] = df_inter["b_poly"]**2 -4*df_inter["AS_squared"]*df_inter["c_poly"]
        
        df_inter["x_ion_1"] = (-df_inter["b_poly"]+np.sqrt(df_inter["Delta"]))/(2*df_inter["AS_squared"]) + pos_antena[0]
        df_inter["y_ion_1"] = df_inter["Yas"]/df_inter["Xas"]*(df_inter["x_ion_1"]-pos_antena[0])+pos_antena[1]
        df_inter["z_ion_1"] = df_inter["Zas"]/df_inter["Xas"]*(df_inter["x_ion_1"]-pos_antena[0])+pos_antena[2] 
        
        df_inter["x_ion_2"] = (-df_inter["b_poly"]-np.sqrt(df_inter["Delta"]))/(2*df_inter["AS_squared"]) + pos_antena[0]
        df_inter["y_ion_2"] = df_inter["Yas"]/df_inter["Xas"]*(df_inter["x_ion_2"]-pos_antena[0])+pos_antena[1]
        df_inter["z_ion_2"] = df_inter["Zas"]/df_inter["Xas"]*(df_inter["x_ion_2"]-pos_antena[0])+pos_antena[2]
        
        df_inter["X1xposAntena"] = (df_inter["x_ion_1"]-pos_antena[0])*df_inter["Xas"] + (df_inter["y_ion_1"]-pos_antena[1])*df_inter["Yas"] + (df_inter["z_ion_1"]-pos_antena[2])*df_inter["Zas"]
        df_inter["X2xposAntena"] = (df_inter["x_ion_2"]-pos_antena[0])*df_inter["Xas"] + (df_inter["y_ion_2"]-pos_antena[1])*df_inter["Yas"] + (df_inter["z_ion_2"]-pos_antena[2])*df_inter["Zas"]

        
        condition = (df_inter["X1xposAntena"]<0) | ( (df_inter["X1xposAntena"]>0) & (df_inter["X2xposAntena"]>0) &  (df_inter["X2xposAntena"]<df_inter["X1xposAntena"]))
        
        df_inter["pos_ion_X"] = np.where(condition,df_inter["x_ion_2"],df_inter["x_ion_1"]) 
        df_inter["pos_ion_Y"] = np.where(condition,df_inter["y_ion_2"],df_inter["y_ion_1"]) 
        df_inter["pos_ion_Z"] = np.where(condition,df_inter["z_ion_2"],df_inter["z_ion_1"]) 

        df["lat"],df["lon"],df["alt"] = pm.ecef2geodetic(df_inter["pos_ion_X"],df_inter["pos_ion_Y"],df_inter["pos_ion_Z"])
        return df

        #return 1.0, [20000000,2000000,-1000000]
        #pos_sat=self.getPosdf(df)
    def getElevation_old(self,sat,pos_antena,date,mode="RAD"):   
        n = np.array([pos_antena[0],pos_antena[1],pos_antena[2]])
        v = np.array([pos_sat[0]-pos_antena[0],pos_sat[1]-pos_antena[1],pos_sat[2]-pos_antena[2]])
        nv = np.dot(n,v)
        sin_phi = nv/(np.linalg.norm(n)*np.linalg.norm(v))
        return math.asin(sin_phi),pos_sat


    # Read position from saved files
    # Not working but not used
    def getPos(self,df):
        
        print (self.df_pos)
        print (df)
        print (pd.merge(df,self.df_pos,how='left',on=['sv','time']))
        sys.exit()
        
        if len(self.dict_df_pos[sat])==0: 
            return [float("NaN"),float("NaN"),float("NaN")]

        if date not in self.dict_df_pos[sat].index: 
            return [float("NaN"),float("NaN"),float("NaN")]

        pos = self.dict_df_pos[sat][["X","Y","Z"]].loc[date].tolist()
       
        return pos  

          
    #### Old function, should not be used anymore
    def to_CSV(self,date,feather=True):
        #resetCSV(date,feather)
        #print ("In resetcsv",gps_dir)
                
        year_gpsdir = self.gps_dir + str(date.year)
        if not os.path.exists(year_gpsdir):
            try: os.mkdir(year_gpsdir)
            except OSError as e:
                if e.errno!=17: print ("FAIL creation of directory "+year_gpsdir, e )
                else: print ("Successfully created the directory "+year_gpsdir)
                
        #Annex file reporting doy that have been considered for each satellite
        f_doy_reported = self.gps_dir + str(date.year) + "/reported_doy.feather"
        if not os.path.exists(f_doy_reported):
            dict_doy_reported = {"sat":[]}
            for doy in range(366):
                dict_doy_reported[str(doy)] = []
            for i in range(1,33):
                #Name of the satellite
                sat = ""
                if i<10: sat = "G0"+str(i)
                else: sat = "G"+str(i)
                dict_doy_reported["sat"].append(sat)
                
                for doy in range(366):
                    dict_doy_reported[str(doy)].append(0)
            pd.DataFrame(dict_doy_reported).to_feather(f_doy_reported)
        
        
        df_doy_reported = pd.read_feather(f_doy_reported)
        df_doy_reported.set_index("sat",inplace=True)
	
        print (df_doy_reported)

        try:
            nav = gr.load(self.f_rinex_nav)
            df_nav = nav.to_dataframe()
        except ValueError:
            print ("COuld not load file", self.f_rinex_nav)
        
        df_nav.index.set_names(["time","sv"],inplace=True)
        df_nav.reset_index(level=["sv"],inplace=True)
        df_nav.dropna(inplace=True)
         
        list_cols = ["Toe","TGD","IDOT","IODC","GPSWeek","TransTime","SVclockBias","SVclockDrift",\
        "SVclockDriftRate","sqrtA","Eccentricity","Io","Omega0","omega","M0","DeltaN","OmegaDot","Cus","Cuc",\
        "Cis","Crs","Crc","Cic"]
         
        for i in range(1,33):
            #Name of the satellite
            sat = ""
            if i<10: sat = "G0"+str(i)
            else: sat = "G"+str(i)

            is_doy_reported = df_doy_reported[str(self.doy)].loc[sat]==1

            # If this doy for this sat has been reported skip it
            if is_doy_reported: continue
                 
            # import data of this satellite
            df_sat_new = df_nav[df_nav["sv"]==sat][list_cols]
            # If other doys have been reported, we want to add the new doy to them
            if os.path.exists(year_gpsdir+"/"+sat+".feather"):
                df_sat = pd.read_feather(year_gpsdir+"/"+sat+".feather")
                df_sat.index = df_sat["time"]  
                df_sat_new = pd.concat(df_sat_new,df_sat)
            df_sat_new.reset_index(inplace=True)
            df_sat_new.to_feather(year_gpsdir+"/"+sat+".feather")
            
            # Report that this doy has been treated to no repeat
            df_doy_reported[str(self.doy)].loc[sat]=1
        df_doy_reported.reset_index(inplace=True)
        df_doy_reported.to_feather(f_doy_reported)

    #### Old function, should not be used anymore
    def getPos2(self,sat,date):
        doy = str(date.timetuple().tm_yday)
        year = str(date.year)
        
        # File correponding to satellite of interest
        csv_nav_sat = st.root_dir + "GPS/" + str(date.year) + "/" + sat + ".feather"
        
        # Check if feather file is not already there in order not to use navigation file
        if not os.path.exists(csv_nav_sat):
            print (csv_nav_sat, "does not exist")
            self.to_CSV(date)
        df_sat_nav = pd.read_feather(csv_nav_sat) 
        df_sat_nav.set_index("time",inplace=True)
        df_sat_nav.index = pd.to_datetime(df_sat_nav.index)
         
        dt=datetime.timedelta(days=2)
        closest_t = date
        
        if len(df_sat_nav)==0:
            print ("Position of sat ", sat, "unknown at",date )
            return [float("NaN"),float("NaN"),float("NaN")]
        
        for t in df_sat_nav.index:
            # Test for dates before time seeked
            #if d<t:
            if (t-date)<dt:
                closest_t = t
                dt = t-date
                
        #print (sat,t)
        data = df_sat_nav.loc[t]    

        return gps_nav_to_XYZ(data,date)    
        



    # return elevation of satellite given the ID os the satellite, the position of the antenna.
    def getElevation_old(self,sat,pos_antena,date,mode="RAD"):
        #return 1.0, [20000000,2000000,-1000000]
        pos_sat=self.getPos(sat,date)
        n = np.array([pos_antena[0],pos_antena[1],pos_antena[2]])
        v = np.array([pos_sat[0]-pos_antena[0],pos_sat[1]-pos_antena[1],pos_sat[2]-pos_antena[2]])
        nv = np.dot(n,v)
        sin_phi = nv/(np.linalg.norm(n)*np.linalg.norm(v))
        return math.asin(sin_phi),pos_sat


def get_pos_antenna(station,year,doy):
    df_antennas = pd.read_csv(st.root_dir+str(year)+"/"+str(doy)+"/antennas.csv")
    df_antennas.set_index("station",inplace=True)
    pos_antena = []
    pos_antena.append(df_antennas.loc[station,"x"])
    pos_antena.append(df_antennas.loc[station,"y"])
    pos_antena.append(df_antennas.loc[station,"z"])
    return pos_antena

def getPiercingPoint_old(pos_antena,pos_sat,h=400000):
    R_I = R_E + h
    #### Calculation of the position of the intersection of Satellite-Antena line with ionosphere which is a second degree equation
    Xas = pos_sat[0]-pos_antena[0]
    # return elevation of satellite given the ID os the satellite, the position of the antenna.
    Yas = pos_sat[1]-pos_antena[1]
    Zas = pos_sat[2]-pos_antena[2]
    AS_squared = Xas**2 + Yas**2 + Zas**2
    #Parameters of the polynomial to solve
    a_poly = AS_squared
    b_poly = 2*Xas*(pos_antena[0]*Xas + pos_antena[1]*Yas + pos_antena[2]*Zas)
    c_poly = Xas**2 * ( pos_antena[0]**2 + pos_antena[1]**2+ pos_antena[2]**2 - R_I**2 )
    Delta = b_poly**2 - 4*a_poly * c_poly
    x_ion, y_ion, z_ion = float('nan'),float('nan'),float('nan')
    pos_ion = [float('nan'),float('nan'),float('nan')]
    if Delta > 0:
        x_ion_1 = (-b_poly+np.sqrt(Delta))/(2*a_poly) + pos_antena[0]
        y_ion_1 = Yas/Xas*(x_ion_1-pos_antena[0])+pos_antena[1]
        z_ion_1 = Zas/Xas*(x_ion_1-pos_antena[0])+pos_antena[2]

        x_ion_2 = (-b_poly-np.sqrt(Delta))/(2*a_poly) + pos_antena[0]
        y_ion_2 = Yas/Xas*(x_ion_2-pos_antena[0])+pos_antena[1]
        z_ion_2 = Zas/Xas*(x_ion_2-pos_antena[0])+pos_antena[2]

        # point must be between satellite and antena, este must be the other solution
        X1xposAntena=(x_ion_1-pos_antena[0])*Xas + (y_ion_1-pos_antena[1])*Yas + (z_ion_1-pos_antena[2])*Zas
        X2xposAntena=(x_ion_2-pos_antena[0])*Xas + (y_ion_2-pos_antena[1])*Yas + (z_ion_2-pos_antena[2])*Zas
        if X1xposAntena<0 or (X1xposAntena>0 and X2xposAntena>0 and X2xposAntena<X1xposAntena): pos_ion = [x_ion_2,y_ion_2,z_ion_2]
        else: pos_ion = [x_ion_1,y_ion_1,z_ion_1]
    elif Delta==0:
        x_ion = -b_poly/(2*a_poly)
        pos_ion = [x_ion,Yas/Xas*(x_ion-pos_antena[0])+pos_antena[1],Zas/Xas*(x_ion-pos_antena[0])+pos_antena[2]]
    return pos_ion

def resetCSV(date,feather=True):
    gps_dir = st.root_dir + "GPS/" #str(date.year)+"/"+ str(date.timetuple().tm_yday) +"/GPS/"
    print ("In resetcsv",gps_dir)
    try: os.mkdir(gps_dir)
    except OSError as e:
        if e.errno!=17: print ("FAIL creation of directory "+gps_dir, e )
    else: print ("Successfully created the directory "+gps_dir)
    year_gpsdir = gps_dir + str(date.year)
    try: os.mkdir(year_gpsdir)
    except OSError as e:
        if e.errno!=17: print ("FAIL creation of directory "+year_gpsdir, e )
    else: print ("Successfully created the directory "+year_gpsdir)
    #for n_sat in range(1,33):
    #    if n_sat<10: sat = "G0" + str(n_sat)
    #    else: sat = "G" + str(n_sat)
    #    sat_dir = gps_dir+sat
    #    try: os.mkdir(sat_dir)
    #    except OSError as e:
    #        if e.errno!=17: print ("FAIL creation of directory "+gps_dir, err)
    #    else: print ("Successfully created the directory "+sat_dir)
    #    for filename in os.listdir(sat_dir):
    #        file_path = os.path.join(sat_dir, filename)
    #        try:
    #            if os.path.isfile(file_path) or os.path.islink(file_path): os.unlink(file_path)
    #            elif os.path.isdir(file_path): shutil.rmtree(file_path)
    #        except Exception as e: print('Failed to delete %s. Reason: %s' % (file_path, e))




def getGPSSatPosition(sat,d,pos_antena=None):
	#csv_nav_dir = dir_rinex_nav + "GPS/" + sat + "/"
	csv_nav_dir = st.root_dir + str(d.year)+"/"+ str(d.timetuple().tm_yday) +"/" + "GPS/" + sat + "/"
	if not os.path.isdir(csv_nav_dir):
		print (f"Calculating RINEX navigation to csv for year:{d.year}, doy:{d.timetuple().tm_yday}")
		makeCSV(d)
		print ("...Done")
	nav_csv_files = [f for f in listdir(csv_nav_dir) if isfile(join(csv_nav_dir, f))]
	if len(nav_csv_files)==0:
		print (f"Calculating RINEX navigation to csv for year:{d.year}, doy:{d.timetuple().tm_yday}")
		makeCSV(d)
		print ("...Done")
		nav_csv_files = [f for f in listdir(csv_nav_dir) if isfile(join(csv_nav_dir, f))]
	csv_nav = csv_nav_dir+nav_csv_files[0]

	df_data = pd.read_csv(csv_nav)
	df_data["time"]=pd.to_datetime(df_data["time"])
	df_data.set_index("time",inplace=True)
	dt=datetime.timedelta(days=2)
	closest_t = d
	for t in df_data.index:
		# Test for dates before time seeked
		#if d<t:
		if (t-d)<dt:
			closest_t = t
			dt = t-d

	data = df_data.loc[t]
	Toe = data['Toe']
	TGD = data['TGD']
	IDOT = data['IDOT']
	IODC = data['IODC']
	GPSWeek = data['GPSWeek']
	TransTime = data['TransTime']
	#clock correction biais
	clockBiais = data['SVclockBias']
	#clock correction drift
	clockDrift = data['SVclockDrift']
	#clock correction drift rate
	clockDriftRate = data['SVclockDriftRate']
	#Square root of the semi-major axis m^{1/2}
	sqrtA =  data['sqrtA']
	#Eccentricity
	Eccentricity =  data['Eccentricity']
	#Inclination angle at reference time
	Io =  data['Io']
	#Longitude of ascending node at reference time
	Omega0 =  data['Omega0']
	#Argument of perigee (semicircles)
	omega = data['omega']
	#Mean anomaly at reference time (semicircles)
	M0 = data['M0']
	#Mean motion difference from computed value
	DeltaN = data['DeltaN']
	#Rate of change of right ascension
	OmegaDot = data['OmegaDot']
	#Rate of change of inclination
	IDOT = data['IDOT']
	#Amplitude of the sine harmonic correction term to the argument of latitude (rad)
	Cus = data['Cus']
	#Amplitude of the cosine harmonic correction term to the argument of latitude (rad)
	Cuc = data['Cuc']
	#Amplitude of the sine harmonic correction term to the angle of inclination (rad)
	Cis = data['Cis']
	#Amplitude of the cosine harmonic correction term to the angle of inclination (rad)
	Cic = data['Cic']
	#Amplitude of the sine harmonic correction term to the orbit radius (m)
	Crs = data['Crs']
	#Amplitude of the cosine harmonic correction term to the orbit radius (m)
	Crc = data['Crc']

	d0 = getDateFromSat(GPSWeek,Toe)

	tk = (d-d0).seconds

	n0 = math.sqrt(GM)/sqrtA**3
	n = n0 + DeltaN

	Mk = M0 + n*tk
	Ek = Mk
	for i in range(3): Ek = Mk + Eccentricity*math.sin(Ek)

	sin_vk = math.sqrt(1-Eccentricity**2)*math.sin(Ek) / (1.0-Eccentricity*math.cos(Ek))
	cos_vk = (math.cos(Ek)-Eccentricity)/(1.0-Eccentricity*math.cos(Ek))
	vk=0
	if sin_vk>0: vk = math.acos(cos_vk)
	else: vk = -math.acos(cos_vk)

	Phik = vk + omega
	delta_uk = Cuc * math.cos(2*Phik) + Cus * math.sin(2*Phik)
	delta_rk = Crc * math.cos(2*Phik) + Crs * math.sin(2*Phik)
	delta_ik = Cic * math.cos(2*Phik) + Cis * math.sin(2*Phik)

	uk = Phik + delta_uk
	rk = sqrtA**2 * (1 - Eccentricity * math.cos(Ek)) + delta_rk
	ik = Io + IDOT*tk + delta_ik

	Xk_p = rk * math.cos(uk)
	Yk_p = rk * math.sin(uk)

	Omegak = Omega0 + (OmegaDot-omega_e)*tk - omega_e * Toe

	Xk = Xk_p * math.cos(Omegak) - Yk_p * math.sin(Omegak)* math.cos(ik)
	Yk = Xk_p * math.sin(Omegak) + Yk_p * math.cos(Omegak)* math.cos(ik)
	Zk = Yk_p * math.sin(ik)
	return [Xk,Yk,Zk]



def getElevation(sat,pos_antena,date,mode="RAD"):
	#return 1.0, [20000000,2000000,-1000000]
	pos_sat=getGPSSatPosition(sat,date,pos_antena)
	n = np.array([pos_antena[0],pos_antena[1],pos_antena[2]])
	v = np.array([pos_sat[0]-pos_antena[0],pos_sat[1]-pos_antena[1],pos_sat[2]-pos_antena[2]])
	nv = np.dot(n,v)
	sin_phi = nv/(np.linalg.norm(n)*np.linalg.norm(v))
	if mode=="DEG": return math.asin(sin_phi)*180.0/math.pi, pos_sat
	return math.asin(sin_phi),pos_sat

def getBias(sat,date,mode="P1P2"):
	if mode=="P1P2": filename = mode+date.strftime("%y%m")+".DCB"
	elif mode=="P1C1": filename = mode+date.strftime("%y%m")+".DCB"
	path_local_file = st.root_dir + str(date.year)+"/"+ str(date.timetuple().tm_yday) +"/" + filename
	if not os.path.exists(path_local_file):
		compressed_file = filename + ".Z"
		url_compressed_file = "http://ftp.aiub.unibe.ch/CODE/" + str(date.year) + "/" + compressed_file
		path_compressed_file = dir_sat_bias + compressed_file
		#print (path_local_file)
		try:
			print ("Download " + compressed_file)
			urllib.request.urlretrieve(url_compressed_file,path_compressed_file)
		except urllib.error.URLError as e:
			print ("Problem in retrieving: " + url_compressed_file +"\n" + e.reason)
			return 0
		with open(path_compressed_file,'rb') as fh:
			f=open(path_local_file,"w")
			f.write(unlzw(fh.read()).decode("utf-8"))
			f.close()
		os.remove(path_compressed_file)
	f = open(path_local_file)
	for lin in f:
		splt_lin = lin.split()
		if len(splt_lin)==0: continue
		if splt_lin[0]!=sat: continue
		else: return float(splt_lin[1])
	print ("No biais found")
	return 0.0

def getBias_fromfile(sat,f_bias):
    path_local_file = f_bias
    f = open(path_local_file)
    for lin in f:
         splt_lin = lin.split()
         if len(splt_lin)==0: continue
         if splt_lin[0]!=sat: continue
         else: return float(splt_lin[1])
    print ("No biais found")
    return 0.0


'''
	return list of dictionnaries, each correspond to information of the arcs in he time series
	each arc is described with keys:
		- "start" : First point of satellite arc
		- "end" : Last point of satellite arc
		- "tmax": time of maximum elevation
		- "full": True if begining and end of arc is contained is the time serie, if truncated, False
	Input: pandas Serie of elevations with time index.
		t_begin and t_end are to indicate time borders of the time serie.
'''
def get_arcs(elevations,t_begin=None,t_end=None):
    elevations = elevations.dropna()
    values = elevations.values
    index = elevations.index

    if t_begin==None: t_begin = index[0]
    if t_end==None: t_end = index[-1]

    # List of dictionaries containing first and last date of arc and if arc is complete in the time serie
    list_arcs=[]

    # Declare first dict, first arc.
    arc={}

    # Variable to track max elevation during the arc.
    max_elevation = 0
    iin=0

    # If first date has data, the first arc is already running
    # First date of arc is first non nan value, and arc not fully available in time serie
    if len(index)==0 or len(values)==0: return list_arcs
    ioldold,voldold = index[0],values[0]
    iold,vold = index[0],values[0]
    i,v = index[0],values[0]
    #isnan1 = math.isnan(v)
    is_in_arc = False
    if v>0:
        if len(index)>1: inew,vnew=index[1],values[1]
        else:  inew,vnew=index[0],values[0]
        arc["start"]=i
        iin=0
        max_elevation=v
        arc["end"]=None
        if len(values)>1:
            if values[1]<values[0]:
                arc["max_ele"]=values[0]
                arc["tmax"]=index[0]
                arc["imax"]=0
            else:
                arc["max_ele"]=values[1]
                arc["tmax"]=index[1]
                arc["imax"]=0

        if i-t_begin<datetime.timedelta(minutes=5) and v<15.0*math.pi/180: arc["full"]=False
        else: arc["full"]=True
        is_in_arc = True



	# Get elevation sense, rising (positive) or falling (negative)
    last_last_rising=False
    last_rising=False
    rising=False

    if len(elevations)==1:
        arc["tmax"]=index[0]
        arc["imax"]=0
        arc["max_ele"]=values[0]
    for k in range(len(elevations)-1):
		# Get new value of elevation
        inew,vnew = index[k+1],values[k+1]
        rising=vnew>v
		#print (inew,vnew,last_rising,rising,is_in_arc)
		#isnan2 = math.isnan(vnew)
        if not is_in_arc and vnew<0:
            ioldold,voldol=iold,vold
            iold,vold=i,v
            i,v=inew,vnew
            last_rising = rising
            continue

        if not is_in_arc and vnew>0:
            #print ("not in arc and vnew>0")
            is_in_arc=True
            arc["start"]=inew
            iin=k
            arc["end"]=None
            arc["tmax"]=i
            arc["imax"]=0
            arc["full"]=True
            max_elevation = v
            arc["max_ele"]=max_elevation

        # if elevation is changing from up to down, max reached
        if k>0 and last_rising and not rising and max_elevation<vnew:
            max_elevation=vnew
            arc["tmax"]=i
            arc["imax"]=k-iin-1
            arc["max_ele"]=max_elevation

		# Case arc is finished when satellite goes down skyline
        if not rising and (vnew<0):
            #print ("going down, skyline passed")
            arc["end"]=i
            list_arcs.append(arc)
            is_in_arc=False
            arc={} # empty dictionnary waiting for next arc

		# Case time serie between falling and next rising is truncated
        if k>1 and  not last_last_rising and rising and (i-iold>datetime.timedelta(hours=1)):
            arc["end"]=iold#inew
            list_arcs.append(arc)
            arc={}
            arc["start"]=i
            iin=k
            arc["end"]=None
            arc["tmax"]=inew
            arc["imax"]=0
            arc["full"]=True
            max_elevation = vnew
            arc["max_ele"]=max_elevation
            is_in_arc=True
			#print ("GAP")
			#print (i,iold,i-iold)
			#print (list_arcs)

        if k>0 and not last_rising and rising and (v<0):
            arc["end"]=i
            list_arcs.append(arc)
            arc={}
            arc["start"]=i
            iin=k
            arc["end"]=None
            arc["tmax"]=i
            arc["imax"]=0
            arc["full"]=True
            max_elevation = v
            arc["max_ele"]=max_elevation
            is_in_arc=True
			#print ("CHANGING ARC v<0")
			#print (i,iold,i-iold)
			#print (list_arcs)

        iold,vold=i,v
        i,v=inew,vnew

        last_last_rising = last_rising
        last_rising = rising


	# Special treatment for last data of time series, if an arc was running end it and set no "full" arc
    if is_in_arc:
        arc["end"]=inew
        if rising:
            arc["tmax"]=inew
            arc["imax"]=k-iin
        else: arc["full"]=True
        arc["max_ele"]=elevations[arc["tmax"]]
        list_arcs.append(arc)

	#for arc in list_arcs: print (arc)
    return list_arcs



#############################################################################
#
# MIT License
#
# Copyright (c) 2021 Sylvain Blunier
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#
#############################################################################

#import gnss
import georinex as gr
#import stations as st
import pandas as pd
import scipy.constants as csts
import datetime
import numpy as np
import sys,os,subprocess
#import matplotlib.pyplot as plt
#import matplotlib.dates as md
import pymap3d as pm
#from os import listdir,path,mkdir
#from os.path import isfile, join
import julian
import math	
import time

#from .stations import root_dir
from . import stations as st
from . import gnss



pd.options.mode.chained_assignment = None

# Earth Radius
R_E = 6371000

## Definition of the constants for the equations of STEC
f1,f2 = 1575.42 * 1e6, 1227.60 * 1e6
lambda1,lambda2=csts.c/f1,csts.c/f2
alpha = f1**2*f2**2/(f1**2-f2**2)/40.318

# List of possible indices in observation RINEX files, P1 not available
#full_vars_list = ['P1','P2','C1','C2','L1','L2','S1','S2']
full_vars_list = ['P2','C1','C2','L1','L2','S1','S2']

def fit_lin(t,sig):
    N=len(t)
    # Coefficients of paraboloid of error function (N*mse)
    a,b,c,d,e=0,N,0,0,0 # b=N for B^2 coef
    # Iterate over subseries to calculate coefficients
    for i in range(N):
        a+=t[i]**2 # A^2 coef
        c+=2*t[i] # A*B coef
        d-=2*t[i]*sig[i] # A coef
        e-=2*sig[i] # B coef

    # Forward A and B parameters of linear fit (solve the minimum of mse)
    A=-(2*b*d-c*e)/(4*a*b-c**2)
    B=-(c*d-2*a*e)/(c**2-4*a*b)

    sigma=0
    sigmas = []
    # Calculate value of err function
    for i in range(N):
        s = abs(sig[i]-A*t[i]-B)
        sigma+=s
        sigmas.append(s)
    max_deviation = max(sigmas)
    mean_deviation = np.mean(sigmas)

    return A,B,max_deviation,mean_deviation

def tleft(border):
    return border["t_left"]

#def filter_overlap_leap(list_borders,series):
#    return list_borders

def filter_slope_leap(list_borders,series,diffs):
    list_borders.sort(key=tleft)
    #return list_borders
    for i in range(len(list_borders)-1):
        if list_borders[i+1]["left"]<=list_borders[i]["right"]:
            list_borders[i+1]["left"]=list_borders[i]["right"]+1
            list_borders[i+1]["t_left"]=series.index[list_borders[i]["right"]+1]
    i=0
    while i<len(list_borders):
        #short segment
        t_short_segment = abs((list_borders[i]["t_right"]-list_borders[i]["t_left"]).seconds)<3*60
        short_segment = list_borders[i]["right"]-list_borders[i]["left"]<6
        negative_segment = list_borders[i]["t_right"]<list_borders[i]["t_left"]
        if t_short_segment or negative_segment or short_segment:
            list_borders.pop(i)
        else: i+=1
    i=1
    while i<len(list_borders)-1:
        if (list_borders[i]["t_right"]-list_borders[i]["t_left"]).seconds>30*60:
            i+=1
            continue
        A,B,M,m=fit_lin(diffs[list_borders[i]["left"]:list_borders[i]["right"]],series[list_borders[i]["left"]:list_borders[i]["right"]].values)
        if not (list_borders[i-1]["right_A"] is None or list_borders[i+1]["left_A"] is None):
            if abs(A-list_borders[i-1]["right_A"])>0.02 and abs(list_borders[i+1]["left_A"]-A)>0.02:
                list_borders.pop(i)
            else: i+=1
        else:
            i+=1
    return list_borders


def plot_leap(diffs,series,s,A,B,N,borders,title=""):
    print ("plot_leap")
    fig, ax = plt.subplots(1,figsize=(10,7))
    xx = np.array(diffs[s:s+N])
    yy = A*xx+B
    if s<len(diffs):
        xright=diffs[s+N]
        tx_right = series.index[s+N]
        yright=A*xright+B
        ax.plot([tx_right],yright,'rx')
        arrow_yrightmargin = (series[s+N]-yright)*0.012
        x0 = md.date2num(series.index[s+N])
        ax.arrow(x0,yright+arrow_yrightmargin,x0-x0,series[s+N]-yright-2*arrow_yrightmargin,width=0.00003,head_length=0.07*(series[s+N]-yright-2*arrow_yrightmargin),length_includes_head=True,color="black")
    if s>0:
        xleft=diffs[s-1]
        tx_left = series.index[s-1]
        yleft=A*xleft+B
        ax.plot([tx_left],yleft,'rx')
        arrow_yleftmargin = (series[s-1]-yleft)*0.012
        x1 = md.date2num(series.index[s-1])
        ax.arrow(x1,yleft+arrow_yleftmargin,x1-x1,series[s-1]-yleft-2*arrow_yleftmargin,width=0.00003,head_length=0.07*(series[s-1]-yleft-2*arrow_yleftmargin),length_includes_head=True,color="black")

    txx = [series.index[s+i] for i in range(N)]

    ax.plot(txx,yy,'k')
    plt.plot(series.iloc[max(0,s-4):min(s+N+6,len(series))],'x')
    ax.plot(series,'bx')
    for b in borders:
        if b["left"]!=None:
            ax.axvline(x=series.index[b["left"]],color='r')
    for b in borders:
        if b["right"]!=None:
            ax.axvline(x=series.index[b["right"]],color='b')
    #if deviation>tol_deviation: print ("deviation<tol_deviation",deviation,tol_deviation)
    #if deviation>10*max_deviation: print ("deviation>10*max_deviation")
    plt.title(title)
    mng = plt.get_current_fig_manager()
    mng.full_screen_toggle()

    #plt.savefig("cycle_slips_png/almaG01_"+str(n_frame)+".png")
    plt.show()
    plt.close()

from pathlib import Path

class tec:

    def __init__(self,list_f_obs,f_nav,f_sat_bias,outfolder="output",resolution=60,h=400000):
    
        # List of observation files
        self.list_f_obs = list_f_obs        
        # List of navigation files
        self.f_nav = f_nav
        # File containing satellite bias
        self.f_sat_bias = f_sat_bias
        #Output feather file
        #if f_out == "": 
        #    self.f_feather = self.list_f_obs[min(len(self.list_f_obs)-1,1)][:-4]+"tec.feather"
        #else:
        #    if f_out[-8:] ==".feather": self.f_feather = f_out
        #    else: self.f_feather = f_out + ".feather"
        
        if not os.path.isfile(self.f_sat_bias): 
            print ("bias file location not found, will take 0, might affect strongly the results")
    
        # Test if each file in list of stations exists    
        for f_obs in self.list_f_obs:
            if not os.path.exists(f_obs):
                print ("File",f_obs,"does not exist, nothing to process")
                return       
        
        self.station = ""
        self.coord = ""

        # Estimated high of ionosphere
        self.h = h

        #Resolution that will be used for the process (should by 60 seconds)
        self.resolution = 60

        # Resolution of the rinex station, in seconds.
        self.obs_resolution = ""

        # DataFrame containing observation data
        self.df_obs = pd.DataFrame()

        # List of satellites that have been observed
        self.sv = []

        # List of dictionnaries containing borders of signal for each satellite
        self.borders={}

        # Time and year of file
        self.year = None
        self.doy = None

        # Bias of the antenna, calculated by method "compute_reveiver_bias"
        # Value stored in csv "stations.csv"
        self.br = 0
        
        if not os.path.exists(st.root_dir + "TEC/"):
            try: os.mkdir(st.root_dir + "TEC/")
            except OSError as e:
                    if e.errno!=17: print ("FAIL creation of directory "+st.root_dir + "TEC/", e )
            else: print ("Successfully created the directory "+st.root_dir + "TEC/")

    def rinex_to_stec(self):
        ''' Extract the relevant data for tec calculation from observation and navigation
            Compute STEC of pseudo range and code phase
        '''

        # Header for observation
        try: 
            # Header of the central file, the one we want to save the information
            header = gr.rinexheader(self.list_f_obs[min(len(self.list_f_obs)-1,1)])
            # Header of the first file, to extract 
        except ValueError: return False
        self.station =header["MARKER NAME"].replace(" ","").lower()
        self.coord = header['position']
        if "INTERVAL" in header.keys():
            self.obs_resolution = float(header["INTERVAL"].replace(" ",""))
        else:
            self.obs_resolution = 1.0

        t_obs_splt = header["TIME OF FIRST OBS"].split(" ")
        t_obs_list = []
        for t in t_obs_splt:
            if t not in ["","GPS"]:
                t_obs_list.append(float(t))

        date_observation = datetime.datetime(int(t_obs_list[0]),int(t_obs_list[1]),int(t_obs_list[2]),int(t_obs_list[3]),int(t_obs_list[4]),int(t_obs_list[5]))
        self.year = date_observation.year
        self.doy = str(date_observation.timetuple().tm_yday)
        #sys.exit()

        list_vars = []

        # lista de campos accesibles en el header
        fields = header['fields']

        # Revisamos si las variables están disponibles
        for l in full_vars_list:
            if l in fields: list_vars.append(l)

        # Concat all data in one file
        first=True
        self.df_obs = pd.DataFrame()
        for f_obs in self.list_f_obs:
            # Read data of observation file
            obs=gr.load(f_obs,meas=list_vars,use="G")
 
            if first: 
                # Rinex a DataFrame
                self.df_obs = obs.to_dataframe()
                self.df_obs.index.set_names(["time","sv"],inplace=True)
                self.df_obs.reset_index(level=["sv"],inplace=True)
                self.df_obs.index = pd.to_datetime(self.df_obs.index)
                first=False
            else:
                # Rinex a DataFrame
                df_day = obs.to_dataframe()
                df_day.index.set_names(["time","sv"],inplace=True)
                df_day.reset_index(level=["sv"],inplace=True)
                df_day.index = pd.to_datetime(df_day.index)
                self.df_obs = pd.concat([df_day,self.df_obs])
            
            

        # List satellites seen by the station and prepare dict of list_borders
        for s in self.df_obs["sv"].values:
            if s not in self.sv:
                self.sv.append(s)
                self.borders[s]=[]

        # Process Slant TEC for pseudo range (slp)
        self.df_obs['STEC_slp'] = (self.df_obs['P2'] - self.df_obs['C1'])*alpha/1e16
        # Process Slant TEC for code phase (sll)
        self.df_obs['STEC_sll'] = (lambda1*self.df_obs['L1'] - lambda2*self.df_obs['L2'])*alpha/1e16

        # Changing resolution
        self.df_obs = self.df_obs.groupby("sv").resample(str(self.resolution)+"S").mean()
        self.df_obs.reset_index(level=["sv"],inplace=True)

        # Remove nan values (may be optimized)
        self.df_obs.dropna(subset=["STEC_slp","STEC_sll"],inplace=True)

        ## When intermediate feather not used anymore .reset_index() should be removed
        self.df_obs = self.df_obs[["sv","STEC_slp","STEC_sll"]]#.reset_index()
        return True


    def add_satellite_pos(self):
        #time_list =  self.df_obs.index
        gps = gnss.gnss(self.f_nav)
        
        d_in, d_out = min(self.df_obs.index), max(self.df_obs.index)
        gps.load_sats(d_in,d_out)
        df_data = pd.DataFrame()

        t_tot = 0
        t_full = 0
        tf = time.time()
        
        
        self.df_obs = gps.getElevation(self.df_obs,self.coord)
        self.df_obs = gps.getPiercingPoint(self.df_obs,self.coord,self.h)
        
        return True


    def list_leaps_series(self,series,tol_dev=0.2,tol_sig=10,N_=5,debug=False):
        ''' detects leaps in the STEC_sl time series
            input: Data pandas Series indexes with time
            output: list of dictionnaries
                Each dictionary contains first and last date of time series without leap
        '''
        #series = self.df_obs[self.df_obs["sv"]==sat][signal].loc[start:end]

        #Remove nan data in series
        series.dropna(inplace=True)
        indices = series.index
        # Discard too short series
        if len(series)<=N_: return []

        # Final list that will contain all the borders
        list_borders = []
        N=N_

        tol_deviation=tol_dev*self.resolution


        # List time in seconds from first index of series
        diffs = [0]
        for i in range(len(series)-1):
            diffs.append(diffs[-1]+(series.index[i+1]-series.index[i]).seconds)

        s=0

        list_fit_params = {}

        unstable_left=True

        while s<len(series)-N:

            #Going for N next point without big jump
            if False:
                has_big_time_jump=True
                while has_big_time_jump and s<len(series)-N:
                    for i in range(N-1):
                        if diffs[s+1+i]-diffs[s+i]>2*60:
                            print ("TIME JUMP")
                            if not border["left"] is None:
                                border["right"]=s+i
                                list_borders.append(border)
                                border={"left":None,"right":None,"left_A":None,"left_B":None,"right_A":None,"right_B":None}
                                unstable_left=True
                            s=s+1
                            break
                        has_big_time_jump=False

            #Get linear fit parameter of N right points
            A,B,max_dev,mean_dev = fit_lin(diffs[s:s+N],series[s:s+N].values)
            # Compute distance of point s and s+N+1 with fit
            left_dev=abs(series[s-1]-A*diffs[s-1]-B) if s>0 else None
            right_dev=abs(series[s+N]-A*diffs[s+N]-B) if s+N<len(series) else None
            list_fit_params[s]={"A":A,"B":B,"max_dev":max_dev,"left_dev":left_dev,"right_dev":right_dev}
            s+=1


        border={"left":None,"t_left":None,"right":None,"t_right":None,"left_A":None,"left_B":None,"right_A":None,"right_B":None}
        #print (list_fit_params)

        retroactive_borders = []

        s1=0
        s2=0
        while s1<len(series):
            # Looking for an right border
            if border["right"] is None:
                if s1>=len(series)-N+1:
                    #print ("plot 0")
                    #list_borders = remove_overlaps(list_borders)
                    if debug:
                        print ("RIGHT",s1,series.index[s1],len(series),list_fit_params[s1]["left_dev"],list_fit_params[s1]["right_dev"],tol_dev)
                        plot_leap(diffs,series,s1-N,list_fit_params[s1-N]["A"],list_fit_params[s1-N]["B"],N,list_borders,"forward")
                    list_borders = filter_slope_leap(list_borders,series,diffs)
                    return list_borders
                    #return list_borders
                if s1==len(series)-N:
                    border["right"]=s1+N-1
                    border["t_right"]=series.index[s1+N-1]
                    border["right_A"]=list_fit_params[s1-1]["A"]
                    border["right_B"]=list_fit_params[s1-1]["B"]

                    s2=s1-1
                elif list_fit_params[s1]["right_dev"]>tol_deviation or (list_fit_params[s1]["right_dev"]>tol_sig*list_fit_params[s1]["max_dev"] and list_fit_params[s1]["right_dev"]>0.04*self.resolution):
                    if debug:
                        if s1>=0:
                            print ("RIGHT",s1,series.index[s1],len(series),list_fit_params[s1]["left_dev"],list_fit_params[s1]["right_dev"],tol_dev)
                            plot_leap(diffs,series,s1,list_fit_params[s1]["A"],list_fit_params[s1]["B"],N,list_borders)
                    border["right"]=s1+N-1
                    border["t_right"]=series.index[s1+N-1]
                    if s1!=0:
                        border["right_A"]=list_fit_params[s1]["A"]
                        border["right_B"]=list_fit_params[s1]["B"]
                    #print ("plot 2")
                    s2=s1
                else:
                    if debug:
                        if s1>=0: print ("RIGHT",s1-N,series.index[s1],len(series),list_fit_params[s1]["left_dev"],list_fit_params[s1]["right_dev"],tol_dev)
                    s1+=1
            else:
                #print (s1,border)
                if debug:
                    print ("LEFT",s2,series.index[s2],len(series),list_fit_params[s2]["left_dev"],list_fit_params[s2]["right_dev"],tol_dev)
                    plot_leap(diffs,series,s2,list_fit_params[s2]["A"],list_fit_params[s2]["B"],N,list_borders)
                ## Conditions for left border
                # s2 cannot be negative, if at 0 this is a begining of segment
                s2_null = s2==0
                # If s2 reached right border of segment found at its left, must stop
                s2_reached_last_segment=False
                if len(list_borders)>0: s2_reached_last_segment = s2==list_borders[-1]["right"]
                # Conditions based on tolerance threshold
                s2_above_tolerance = False
                s2_std = False
                if list_fit_params[s2]["left_dev"]!=None:
                    # Next point is clearly above tolerance
                    s2_above_tolerance = list_fit_params[s2]["left_dev"]>tol_deviation
                    # Filter by standard deviation of linear left segment
                    s2_std = list_fit_params[s2]["left_dev"]>tol_sig*list_fit_params[s2]["max_dev"] and list_fit_params[s2]["left_dev"]>0.04*self.resolution

                if s2_null or s2_reached_last_segment or s2_above_tolerance or s2_std:
                    #print (border["right"]-s2)
                    if border["right"]-s2<=N:
                        if len(retroactive_borders)!=0:
                            for b in retroactive_borders:
                                #print ("Adding border to list_borders 1",b)
                                list_borders.append(b)
                            retroactive_borders=[]
                            #if debug: plot_leap(diffs,series,s2,list_fit_params[s2]["A"],list_fit_params[s2]["B"],N,list_borders,"forward")

                        border={"left":None,"t_left":None,"right":None,"t_right":None,"left_A":None,"left_B":None,"right_A":None,"right_B":None}
                        s1=s1+N
                        continue
                    border["left"]=s2
                    border["t_left"]=series.index[s2]
                    border["left_A"]=list_fit_params[s2]["A"]
                    border["left_B"]=list_fit_params[s2]["B"]
                    #print ("new", border)
                    retroactive_borders.append(border)
                    # If right border of last segment is not reached, must keep
                    # searching backward for a new segment
                    if len(list_borders)>0 and border["left"]-list_borders[-1]["right"]>N:
                        if s2-1>=N:
                            border={"left":None,"t_left":None,"right":s2-1,"t_right":series.index[s2-1],"left_A":None,"left_B":None,"right_A":list_fit_params[s2-N]["A"],"right_B":list_fit_params[s2-N]["B"]}
                        else:
                            border={"left":None,"t_left":None,"right":s2-1,"t_right":series.index[s2-1],"left_A":None,"left_B":None,"right_A":None,"right_B":None}
                        s2-=N
                        continue
                    retroactive_borders.reverse()
                    for b in retroactive_borders:
                        list_borders.append(b)
                    retroactive_borders=[]
                    #print ("plot 3")
                    #if debug: plot_leap(diffs,series,s2,list_fit_params[s2]["A"],list_fit_params[s2]["B"],N,list_borders,"forward")

                    if s1>=len(series)-N-1:
                        #print (list_borders)
                        #list_borders = remove_overlaps(list_borders)
                        if debug:
                            plot_leap(diffs,series,s2,list_fit_params[s2]["A"],list_fit_params[s2]["B"],N,list_borders)
                        list_borders = filter_slope_leap(list_borders,series,diffs)
                        return list_borders
                    border={"left":None,"t_left":None,"right":None,"t_right":None,"left_A":None,"left_B":None,"right_A":None,"right_B":None}
                    s1=s1+N
                else: s2-=1
        list_borders = filter_slope_leap(list_borders,series,diffs)
        return list_borders

    def list_leaps(self,df):
        borders=[]
        borders_sl = self.list_leaps_series(df["STEC_sll"],0.4,3,3,debug=False)
        borders_slp = self.list_leaps_series(df["STEC_slp"],100,150,5,debug=False)

        # We look at the borders found for slp and sll leaps, and make the
        # match
        # For each border found for sll
        if len(borders_sl)==0 or len(borders_slp)==0: return borders
        for bsll in borders_sl:
            # For each border found for slp
            for bslp in borders_slp:
                # Test if left border of sll in inside slp borders item
                if bsll["t_left"]>=bslp["t_left"] and bsll["t_left"]<bslp["t_right"]:
                    # Left of new border must be the one of bsll
                    newb = bsll.copy()
                    newb["t_left"]=bsll["t_left"]
                    # We asign to the new border the earliest right border
                    if bsll["t_right"]<bslp["t_right"]: newb["t_right"]=bsll["t_right"]
                    else: newb["t_right"]=bslp["t_right"]
                    borders.append(newb)
                    break
                # If last condition not fulfilled, test in the right sll border
                # is inside the slp border item
                elif bsll["t_right"]>bslp["t_left"] and bsll["t_right"]<=bslp["t_left"]:
                    newb = bsll.copy()
                    newb["t_right"]=bsll["t_right"]
                    #Left border must be outside blsp, otherwise it would
                    #have been seen by the previous condition
                    newb["t_left"]=bslp["t_left"]
                    borders.append(newb)
                    break
                # Case the bsll fully contains bslp, take bslp borders
                elif bsll["t_left"]<bslp["t_left"] and bsll["t_right"]>bslp["t_right"]:
                    borders.append(bslp)
                    break

        return borders

    def add_baseline(self):

        df_data = self.df_obs.copy()
        if len(df_data)==0: return

        t_begin = df_data.index[0]
        t_end = df_data.index[-1]

        df_filtered = pd.DataFrame()

        for sat in self.sv:

            df_sat = df_data[df_data["sv"]==sat]
            if len(df_sat)==0: continue

            # Make list of arcs of satellite using elevation information
            elevations = df_sat["elevation"]
            list_arcs = gnss.get_arcs(elevations,t_begin,t_end)

            # Get satellite bias and correct STEC_sl with it
            sat_bias = gnss.getBias_fromfile(sat,self.f_sat_bias) * alpha * csts.c * 1e-9  / 1e16
            df_sat['STEC_sl']=df_sat['STEC_sll']

            # Squared of sin of elevation for baseline (Brs) pondering
            df_sat["sin2_ele"] = np.sin(df_sat['elevation'])**2

            # Individual values of the sum for future baseline calculation
            df_sat['BRs'] = (df_sat['STEC_slp']-df_sat['STEC_sll'])*df_sat["sin2_ele"]

            # cos(chi) to calculate VTEC from STEC
            df_sat['cos_chi'] = np.cos(np.arcsin(R_E*np.cos(df_sat["elevation"])/(R_E+self.h)))

            n_arc=0
            # Run over each arc
            for arc in list_arcs:
                # Must gets borders that are inside the arc segment
                arc_borders = None
                arc_borders = []

                # Extract data from satellite arc
                df_arc = df_sat.loc[arc["start"]:arc["end"]]
                if len(df_arc)<=8:continue

                arc_borders = self.list_leaps(df_arc)

                if len(arc_borders)==0: continue

                # Eliminate segments with too low elevation satellites
                list_max_ele = []
                #for i,b in enumerate(arc_borders):
                i=0
                while i<len(arc_borders):
                    t_dist=(arc_borders[i]["t_right"]-arc_borders[i]["t_left"]).seconds
                    max_ele = max(df_arc.loc[arc_borders[i]["t_left"]:arc_borders[i]["t_right"]]["elevation"])
                    if max_ele<30*np.pi/180 and t_dist<20*60:
                        arc_borders.pop(i)
                    else:
                        i+=1
                        list_max_ele.append(max_ele)

                #### Clasify segments #####
                d_df_seg = []
                has_sane_parts = False
                sane_indices = []
                list_size_segs = []
                for i,b in enumerate(arc_borders):
                    size_seg = (arc_borders[i]["t_right"]-arc_borders[i]["t_left"]).seconds
                    list_size_segs.append(size_seg)

                    if size_seg>50*60:
                        df_seg = None
                        df_seg = df_arc.loc[b["t_left"]:b["t_right"]]
                        df_br = df_seg.dropna(subset=["BRs","sin2_ele"])
                        brs=df_br['BRs'].sum(skipna=True)/df_br["sin2_ele"].sum(skipna=True)
                        df_seg["STEC_sl"] = df_seg["STEC_sl"] + brs
                        d_df_seg.append(df_seg)
                        has_sane_parts = True
                        sane_indices.append(i)
                    else:
                        d_df_seg.append(None)
                if not has_sane_parts: continue

                # paste left part of arc
                for i in range(sane_indices[0]-1,-1,-1):
                    df_seg = None
                    df_seg = df_arc.loc[arc_borders[i]["t_left"]:arc_borders[i]["t_right"]]
                    t_dist=(arc_borders[i+1]["t_left"]-arc_borders[i]["t_right"]).seconds
                    slope = (arc_borders[i+1]["left_A"]+arc_borders[i]["right_A"])/2

                    #df_filter_arc = df_arc.loc[arc_borders[i]["t_left"]:arc_borders[i]["t_right"]]
                    glue=d_df_seg[i+1]["STEC_sl"].iloc[0]-df_seg["STEC_sl"].iloc[-1]-slope*t_dist
                    df_br = df_seg.dropna(subset=["BRs","sin2_ele"])
                    brs=df_br['BRs'].sum(skipna=True)/df_br["sin2_ele"].sum(skipna=True)

                    if abs(glue-brs)>20: correction = brs
                    else: correction = glue
                    df_seg["STEC_sl"] = df_seg["STEC_sl"] + correction
                    d_df_seg[i] = df_seg

                # paste right part of arc
                for i in range(sane_indices[-1]+1,len(arc_borders)):
                    df_seg = None
                    df_seg = df_arc.loc[arc_borders[i]["t_left"]:arc_borders[i]["t_right"]]
                    t_dist=(arc_borders[i]["t_left"]-arc_borders[i-1]["t_right"]).seconds

                    slope = arc_borders[i-1]["right_A"]

                    glue=d_df_seg[i-1]["STEC_sl"].iloc[-1]-df_seg["STEC_sl"].iloc[0]+slope*t_dist
                    df_br = df_seg.dropna(subset=["BRs","sin2_ele"])
                    brs=df_br['BRs'].sum(skipna=True)/df_br["sin2_ele"].sum(skipna=True)
                    if abs(glue-brs)>20: correction = brs
                    else: correction = glue
                    df_seg["STEC_sl"] = df_seg["STEC_sl"] + correction
                    d_df_seg[i] = df_seg


                # Look at intermediate segments
                i = 0
                if len(sane_indices)>=2:
                #if False:
                    for i,s in enumerate(sane_indices[0:len(sane_indices)-1]):
                        ileft = s+1
                        iright = sane_indices[i+1]-1
                        only_right = False
                        only_left = False
                        while ileft<=iright:
                            if not only_left:
                                t_left = (arc_borders[ileft]["t_left"]-arc_borders[ileft-1]["t_right"]).seconds
                                t_right = (arc_borders[iright]["t_left"]-arc_borders[ileft]["t_right"]).seconds
                                if t_left<=t_right or only_right:
                                    df_seg = None
                                    df_seg = df_arc.loc[arc_borders[ileft]["t_left"]:arc_borders[ileft]["t_right"]]
                                    t_dist=(arc_borders[ileft]["t_left"]-arc_borders[ileft-1]["t_right"]).seconds

                                    slope = arc_borders[ileft-1]["right_A"]

                                    glue=d_df_seg[ileft-1]["STEC_sl"].iloc[-1]-df_seg["STEC_sl"].iloc[0]+slope*t_dist
                                    df_br = df_seg.dropna(subset=["BRs","sin2_ele"])
                                    brs=df_br['BRs'].sum(skipna=True)/df_br["sin2_ele"].sum(skipna=True)
                                    if abs(glue-brs)>20: correction = brs
                                    else: correction = glue
                                    df_seg["STEC_sl"] = df_seg["STEC_sl"] + correction

                                    d_df_seg[ileft] = df_seg
                                    ileft+=1
                                elif not only_right:
                                    only_left=True
                                if ileft>iright: break
                            if not only_right:
                                t_left = (arc_borders[iright]["t_left"]-arc_borders[ileft]["t_right"]).seconds
                                t_right = (arc_borders[iright+1]["t_left"]-arc_borders[iright]["t_right"]).seconds
                                if t_left>=t_right or only_left:
                                    df_seg = None
                                    df_seg = df_arc.loc[arc_borders[iright]["t_left"]:arc_borders[iright]["t_right"]]
                                    t_dist=(arc_borders[iright+1]["t_left"]-arc_borders[iright]["t_right"]).seconds

                                    slope = arc_borders[iright+1]["right_A"]

                                    glue=d_df_seg[iright+1]["STEC_sl"].iloc[0]-df_seg["STEC_sl"].iloc[-1]+slope*t_dist
                                    df_br = df_seg.dropna(subset=["BRs","sin2_ele"])
                                    brs=df_br['BRs'].sum(skipna=True)/df_br["sin2_ele"].sum(skipna=True)
                                    if abs(glue-brs)>20: correction = brs
                                    else: correction = glue
                                    df_seg["STEC_sl"] = df_seg["STEC_sl"] + correction

                                    d_df_seg[iright] = df_seg
                                    iright-=1
                                elif not only_left:
                                    only_right=True


                df_filter_arc = pd.DataFrame()
                for df in d_df_seg:
                    if df is None:continue
                    df_filter_arc = pd.concat([df_filter_arc,df])

                # Calculate VTEC (still not take into account receiver bias)
                df_filter_arc["STEC_sl"] = df_filter_arc["STEC_sl"]+sat_bias
                df_filter_arc['VTEC']=df_filter_arc['STEC_sl']*df_filter_arc['cos_chi']
                df_filter_arc=df_filter_arc[["sv","lat","lon","elevation","cos_chi","STEC_sll","STEC_slp","STEC_sl","VTEC"]]
                df_filtered = pd.concat([df_filtered,df_filter_arc])

        self.df_obs = df_filtered.copy()



    def compute_receiver_bias(self):
        ''' Sylvain Blunier 06/2021 v1.0.0
                Function that computes and returns the biais of some station.
                * Input: string of the name of the feather files that contains data for compute
                This file must be in directory "./feather/"
                'station_feather' feather file must at least possess columns:
                    time,elevation,STEC_sl
                * Output: float representing receiver bias
                Algorithm:
                This algorithm minimize the sum of variances between satellites at each time
                *section 4.1 of https://hal.archives-ouvertes.fr/hal-00317176/file/angeo-21-2083-2003.pdf
                *algebra: https://colab.research.google.com/drive/1UCZHR0t-9jyyjAnLuMN3N0Z2NB6tgI_l?usp=sharing
        '''

        df = self.df_obs.copy()

        # Remove rows with nan STEC_sl
        df.dropna(subset=["STEC_sl","elevation"],inplace=True)
        df = df[df["elevation"]>30*np.pi/180]

        # Compute cos\chi for full serie
        df["ci"] = np.cos(np.arcsin(R_E*np.cos(df["elevation"])/(R_E+self.h)))

        # Coefficients of the cuadratic error function that will be computed
        a,b=0,0

        # Create list of time of the station removing duplicates
        time_series = df.groupby(level="time").size()
        time_series = time_series.index

        #print (len(time_series))
        # Compute sums
        for t in time_series:
            sum_ci = 0 # sum of cos\chi_i for a coefficient (to be squared)
            sum_ci2 = 0 # sum of squared of cos\chi_i for a and b coefficient
            sum_sici = 0 # sum of STEC_i*cos\chi_i for b coefficient
            sum_sici2 = 0 # sum of STEC_i*cos\chi_i^2 for b coefficient
            N=0 # Number of satellites with data at time t
            # Get subserie containing data at time t
            df_t = df.loc[t]
            # If only one satellite has data, df_t is a Series, not a Dataframe, nothing to compute
            if isinstance(df_t,pd.Series): continue
            # Compute sums over each satellite
            for index,row in df_t.iterrows():
                si = row["STEC_sl"]
                ci = row["ci"]
                sum_ci += ci
                sum_ci2 += ci**2
                sum_sici += si*ci
                sum_sici2 += si*ci*ci
                N+=1
            if N==0: continue
            # Update a and b over time serie, we ignore the main 1/N factor since it cancels for bias
            # We also ignore "2" factor for a since is cancels in -b/(2a) root calculation
            a=a+(sum_ci2-(1/N)*sum_ci**2)/N
            b=b+(sum_sici2-(1/N)*sum_sici*sum_ci)/N
            #print (a,b,si,ci,sum_ci,sum_ci2,sum_sici,sum_sici2,N)
            if math.isnan(a) or math.isnan(b): sys.exit()

        if a==0: return float("nan")
        # root of error function = receiver bias.
        br = b/a

        self.br = br

        d = {"station":[self.station],"br":[self.br]}

        f_br = st.root_dir + "TEC/" + str(self.year) + "/receiver_bias.csv"
        if not os.path.exists(st.root_dir + "TEC/"+ str(self.year)):
            try: os.mkdir(st.root_dir + "TEC/"+ str(self.year))
            except OSError as e:
                    if e.errno!=17: print ("FAIL creation of directory "+st.root_dir + "TEC/"+ str(self.year), e )
            else: print ("Successfully created the directory "+st.root_dir + "TEC/"+ str(self.year))
                    
        if os.path.exists(f_br):
            df_br = pd.read_csv(f_br).set_index("station")
            df_br = df_br[df_br.index!=self.station]
            df_br = pd.concat([df_br,pd.DataFrame(d).set_index("station")])
            df_br.to_csv(f_br)
        else:
            df_br = pd.DataFrame(d).set_index("station")
            df_br.to_csv(f_br)


    def get_receiver_bias(self,force_compute=False):
        ''' Function that returns the bias of the station given in argument
                If the information is not in stations.csv just return 0
                    (with warning, receiver bias should not be ignored)
            * Input: - station: string. station name
                    - compute_if_nan: bool, False by default, if True, calls compute_receiver_bias
            * Output: float. bias value in TECU
        '''
        receiver_bias = pd.read_csv(st.root_dir+"stations.csv").set_index("station")

        if not self.station in receiver_bias.index:
            print (f"WARNING must update stations.csv, with {station} not in csv, return br=0")
            return 0

        br = receiver_bias.loc[self.station]["br"]
        if force_compute:
            self.br = self.compute_receiver_bias()
            return self.br
        self.br=br
        return br

    def add_receiver_bias(self):
        if len(self.df_obs)==0: return
        self.compute_receiver_bias()

        if math.isnan(self.br):
            print (f"Warning: will null receiver bias")
            self.br=0

        # Remove rows with nan STEC_sl
        self.df_obs.dropna(subset=["STEC_sl"],inplace=True)

        # Compute VTEC with receiver bias
        self.df_obs["VTEC"]=(self.df_obs["STEC_sl"]-self.br)*np.cos(np.arcsin(R_E*np.cos(self.df_obs["elevation"])/(R_E+self.h)))

    def to_feather(self, f_feather):
        print (f_feather)
        self.df_obs[["sv","lat","lon","elevation","STEC_slp","STEC_sl","VTEC"]].reset_index().to_feather(f_feather)


    def compute_vtec(self):	
 
        print ("Converting code and pseudorange from observation rinex to slant tec")
        if not self.rinex_to_stec(): return
    	
        print ("Calculating satellite position from navigation rinex")
        self.add_satellite_pos()
    
        print ("Calculating baseline to correct Slant TEC")
        self.add_baseline()
    
        print ("Calculating receiver bias, correct Slant TEC, compute VTEC")
        self.add_receiver_bias()
        
        print ("Save to feather:",st.root_dir + "TEC/" + str(self.year) + "/" + self.station + str(self.doy) + "tec.feather")
        self.to_feather(st.root_dir + "TEC/" + str(self.year) + "/" + self.station + str(self.doy) + "tec.feather" )
        #self.to_feather(self.f_feather)
    



#############################################################################3
#
# MIT License
#
# Copyright (c) 2021 Sylvain Blunier
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#
#############################################################################3

import georinex as gr
import pandas as pd
import numpy as np

from os import listdir, getenv, path, makedirs
from os.path import isfile, join

import sys

#import os
#root_dir = os.environ.get('PYTEC_PATH')

from pathlib import Path
# create output dir
base_dir = Path(sys.argv[0]).resolve().parent
target_dir = base_dir / "output/"
target_dir.mkdir(parents=True, exist_ok=True)
root_dir = str(target_dir)+"/"
print ("root_dir", root_dir)

def create_output(folder_name: str) -> Path:
    """
    Create a folder in the same directory as the script that started execution
    (e.g. main.py).
    """
    main_file = Path(sys.argv[0]).resolve()
    base_dir = main_file.parent

    target_dir = base_dir / folder_name
    target_dir.mkdir(parents=True, exist_ok=True)

    return target_dir

#def create_output(base_dir: Path, folder_name: str) -> Path:
#    base_dir = Path(base_dir).resolve()
#    target = base_dir / folder_name
#    target.mkdir(parents=True, exist_ok=True)
#    return target

csv_stations = root_dir+"stations.csv"

def resume_station(year,force=False):
    '''  Function not coordinated by now with only code
    Creates stations.csv that contains information about the navigation
    stations
    '''
    
    print ("LIST stations")
    
    if path.exists(csv_stations) and not force: return
    year_folder = root_dir + str(year)+ "/"
    for f in listdir(year_folder):
        doy = f.replace(".zip","")
        day_folder = year_folder+ doy + "/"
        print (day_folder)

        files = [f for f in listdir(day_folder) if isfile(join(day_folder, f))]
        d = {"station":[],"X":[],"Y":[],"Z":[],"interval":[]}
        for f in files:
            print (f)
            if f[-1]!="o": continue
            station = f[:4]
            if station in d["station"]: continue
            if ((station+str(doy)+"0."+str(year)[2:]+"o" not in files) and
                (station+str(doy)+"1."+str(year)[2:]+"o" not in files)): continue

            rinex = station + str(doy) + "0." + str(year)[2:]

            try: header = gr.rinexheader(day_folder+rinex+"o")
            except ValueError:
                print ("Error in file",f)
            if "INTERVAL" in header.keys(): interval = float(header["INTERVAL"].replace(" ",""))
            else: interval = 1.0
            pos_antena = header['position']
            d["station"].append(station)
            d["X"].append(pos_antena[0])
            d["Y"].append(pos_antena[1])
            d["Z"].append(pos_antena[2])
            d["interval"].append(interval)
            #d["br"].append(float("nan"))
    df = pd.DataFrame(d)
    df.sort_values(by="station",inplace=True)
    df.to_csv(csv_stations,index=False)

def get_closest_stations(p):
    ''' Input: list of three float corresponding to the (X,Y,Z) coordinates
        Output: a list of strings corresponding to the station ordered by the closest to farthest '''
    df = pd.read_csv(csv_stations)
    df.set_index("station",inplace=True)
    #p = df[["X","Y","Z"]].iloc[1]
    #print (p)
    dfdist = df.assign(distance = lambda x: ((x["X"]-p[0])**2+(x["Y"]-p[1])**2+(x["Z"]-p[2])**2))
    dfdist.sort_values(by="distance",inplace=True)
    return dfdist.index.values

def get_station_pos(station):
    '''	Input: string, name of a station
    Output: [X,Y,Z] np.ndarray corresponding to its position in the ECEF reference system '''

    df = pd.read_csv(csv_stations).set_index("station")
    pos = df[["X","Y","Z"]].loc[station].values
    return pos

def get_station_interval(station):

    df = pd.read_csv(csv_stations).set_index("station")
    pos = float(df[["interval"]].loc[station].values)
    return pos

#class BaseClass:
#    def base_method(self) -> str:
#        """
#        Base method.
#        """
#        return "hello from BaseClass"

#    def __call__(self) -> str:
#        return self.base_method()


#def base_function() -> str:
#    """
#    Base function.
#    """
#    return "hello from base function"
